#ifndef BOOST_MARSHAL_CLIENT_HPP
#define BOOST_MARSHAL_CLIENT_HPP

#include <boost/marshal/config.hpp>
#include <boost/marshal/detail/threaded_service.hpp>
#include <boost/marshal/detail/expanding_buffer.hpp>
#include <boost/marshal/call_options.hpp>
#include <boost/marshal/call.hpp>
#include <boost/call_traits.hpp>

namespace boost {
namespace marshal {

///	A client used to communicate with a marshal::registry_server over the network.
/**	\param Registry underlying Registry type used.
	\param Connector underlying Connector type used.
	\todo result_buffer needs to be improved.
	\todo this probably needs to split into marshal::client and marshal:registry_client
*/
template<typename Registry, typename Connector>
class client : public detail::threaded_service
{
public:
	typedef typename Registry::id_type id_type;
	typedef typename Registry::archive_type archive_type;

	///	Initializes the client to communicate with the specified Boost.Asio endpoint.
	client(boost::asio::ip::tcp::endpoint endpoint)
		: result_buffer(BOOST_MARSHAL_HEADER_BUFFER_SIZE),
		marshal_buffer(BOOST_MARSHAL_MARSHAL_BUFFER_SIZE),
		connector(endpoint), socket_(service),
		stream(std::ios::in | std::ios::out | std::ios::binary)
	{
		connector.connect(socket_);
		char OK_buffer[] = "OK";
		socket_.receive(boost::asio::buffer(OK_buffer, sizeof(OK_buffer)));

		start_service_thread();
	}

	///	Makes the specified call using the stored default options.
	void operator()(typename boost::call_traits<id_type>::param_type id,
		const call_base &call)
	{
		operator()(id, call, default_options);
	}

	///	Makes the specified call.
	void operator()(typename boost::call_traits<id_type>::param_type id,
		const call_base &call, const call_options &options)
	{
		std::string header_packet = detail::serialize<
			archive_type, void (id_type, call_options, size_t, uintptr_t)>
			(id, options, call.parameters().size(), (uintptr_t) &call);

		boost::asio::write(socket_, asio::buffer(header_packet), boost::asio::transfer_all());
		if (options.marshal_option > call_options::no_results && options.sync_option == call_options::async)
			socket_.async_read_some(result_buffer.whole(),
									boost::bind(&client::read_result, this,
												boost::asio::placeholders::bytes_transferred, boost::asio::placeholders::error));
		
		char OK_buffer[] = "OK";
		socket_.receive(boost::asio::buffer(OK_buffer, sizeof(OK_buffer)));
		boost::asio::write(socket_, asio::buffer(call.parameters()), boost::asio::transfer_all());

		if (options.marshal_option == call_options::no_results)
			return;
		// some results need to come back
		if (options.sync_option == call_options::sync)
		{
			size_t bytes_transferred = 
				socket_.receive(result_buffer.whole());
			process_result(bytes_transferred);
		}
	}
private:
	void read_result(std::size_t bytes_transferred, const boost::system::error_code& error)
	{
		if (!error)
			process_result(bytes_transferred);
	}
	void process_result(std::size_t bytes_transferred)
	{
		stream.str(std::string(result_buffer, bytes_transferred));
		typename Registry::archive_type::iarchive_type archive(stream);

		typename Registry::id_type id;
		call_options options;
		size_t marshal_packet_size;
		uintptr_t call_ptr;

		archive & id;
		archive & options;
		archive & marshal_packet_size;
		archive & call_ptr;

		if (options.marshal_option == call_options::return_only)
		{
			char OK_buffer[] = "OK";
			socket_.send(boost::asio::buffer(OK_buffer, sizeof(OK_buffer)));

			size_t marshal_size = socket_.receive(boost::asio::buffer(marshal_buffer.whole()));
			((call_base *) call_ptr)->result_string(
				std::string(marshal_buffer, marshal_size), options);
		}
		((call_base *) call_ptr)->complete();
	}

	detail::expanding_buffer result_buffer;
	detail::expanding_buffer marshal_buffer;
	std::stringstream stream;

	call_options default_options;

	boost::asio::ip::tcp::socket socket_;
	Connector connector;
};

} // namespace marshal
} // namespace boost

#endif BOOST_MARSHAL_CLIENT_HPP