#ifndef BOOST_MARSHAL_REGISTRY_SERVER
#define BOOST_MARSHAL_REGISTRY_SERVER

#include <boost/marshal/config.hpp>
#include <boost/marshal/detail/expanding_buffer.hpp>
#include <boost/marshal/registry.hpp>
#include <boost/marshal/call_options.hpp>
#include <boost/shared_ptr.hpp>
#include <sstream>

namespace boost {
namespace marshal {

template<typename Registry>
class registry_server
{
public:
	typedef typename Registry::id_type id_type;
	typedef typename Registry::archive_type archive_type;

	registry_server(Registry &reg, boost::shared_ptr<boost::asio::ip::tcp::socket> sock)
		: header_buffer(BOOST_MARSHAL_HEADER_BUFFER_SIZE),
		marshal_buffer(BOOST_MARSHAL_MARSHAL_BUFFER_SIZE),
		stream(std::ios::in | std::ios::out | std::ios::binary),
		registry_(reg), socket_ptr(sock)
	{
		char OK_buffer[] = "OK";
		start_async_read();
		socket_ptr->send(boost::asio::buffer(OK_buffer, sizeof(OK_buffer)));
	}
	~registry_server()
	{
		delete[] header_buffer;
	}
private:
	void start_async_read()
	{
		socket_ptr->async_read_some(header_buffer.whole(),
			boost::bind(&registry_server::read_header, this,
			boost::asio::placeholders::bytes_transferred, boost::asio::placeholders::error));
	}
	void read_header(std::size_t bytes_transferred, const boost::system::error_code& error)
	{
		header_size = bytes_transferred;
		if (!error)
		{
			stream.str(std::string(header_buffer, bytes_transferred));
			typename archive_type::iarchive_type archive(stream);

			archive & id;
			archive & options;
			archive & marshal_packet_size;
			
			
			boost::asio::async_read(*socket_ptr,
									boost::asio::buffer(marshal_buffer, marshal_packet_size),
									boost::bind(&registry_server::read_marshal, this,
												boost::asio::placeholders::error));
			marshal_buffer.guarantee(marshal_packet_size);
			char OK_buffer[] = "OK";
			socket_ptr->send(boost::asio::buffer(OK_buffer, sizeof(OK_buffer)));
		}
	}
	void read_marshal(const boost::system::error_code& error)
	{
		std::string result;
		start_async_read();
		registry_(id, std::string(marshal_buffer, marshal_packet_size), options, &result);
		if (options.marshal_option >= call_options::completed_only)
			socket_ptr->send(boost::asio::buffer(header_buffer, header_size));
		if (options.marshal_option >= call_options::return_only)
		{
			char OK_buffer[] = "OK";
			socket_ptr->receive(boost::asio::buffer(OK_buffer, sizeof(OK_buffer)));

			socket_ptr->send(boost::asio::buffer(result));
		}
	}

	size_t header_size;
	detail::expanding_buffer header_buffer;
	detail::expanding_buffer marshal_buffer;
	std::stringstream stream;

	Registry &registry_;
	boost::shared_ptr<boost::asio::ip::tcp::socket> socket_ptr;

	call_options options;
	id_type id;
	size_t marshal_packet_size;
};

} // namespace marshal
} // namespace boost

#endif // BOOST_MARSHAL_REGISTRY_SERVER